#include <torch/extension.h>
#include <ATen/ATen.h>
#include <cuda_runtime.h>

// This kernel fuses the elementwise multiplication by a scalar, spatial reduction (global average pooling)
// and accumulation across channels (final mean). Each block handles one (batch, channel) slice.

template <int BLOCK_SIZE>
__global__ void fused_global_avg_kernel(
    const float* __restrict__ input,
    float* __restrict__ global_accum, // pointer to a single float to accumulate all channel averages
    const int N,
    const int C,
    const int H,
    const int W,
    const float multiplier
) {
    extern __shared__ float sdata[];

    int tid = threadIdx.x;
    int bid = blockIdx.x; // each block corresponds to one (batch, channel) combination
    int num_elements = H * W;

    // Determine the batch and channel for this block
    int batch_idx = bid / C;
    int channel_idx = bid % C;

    // Pointer to the (batch_idx, channel_idx) slice of the input
    const float* channel_input = input + (batch_idx * C * num_elements) + (channel_idx * num_elements);

    // Each thread accumulates partial sum for its assigned spatial elements
    float sum = 0.0f;
    for (int i = tid; i < num_elements; i += BLOCK_SIZE) {
        sum += channel_input[i];
    }

    // Store each thread's partial sum in shared memory
    sdata[tid] = sum;
    __syncthreads();

    // Reduction in shared memory
    if (BLOCK_SIZE >= 512) {
        if (tid < 256) { sdata[tid] += sdata[tid + 256]; } 
        __syncthreads();
    }
    if (BLOCK_SIZE >= 256) {
        if (tid < 128) { sdata[tid] += sdata[tid + 128]; } 
        __syncthreads();
    }
    if (BLOCK_SIZE >= 128) {
        if (tid < 64) { sdata[tid] += sdata[tid + 64]; } 
        __syncthreads();
    }

    // Warp-level reduction (no __syncthreads needed within a warp)
    if (tid < 32) {
        volatile float* vsmem = sdata;
        vsmem[tid] += vsmem[tid + 32];
        vsmem[tid] += vsmem[tid + 16];
        vsmem[tid] += vsmem[tid + 8];
        vsmem[tid] += vsmem[tid + 4];
        vsmem[tid] += vsmem[tid + 2];
        vsmem[tid] += vsmem[tid + 1];
    }

    // Thread 0 computes the channel average and adds it atomically to the global accumulator
    if (tid == 0) {
        // Instead of applying multiplier elementwise before reduction, we do it here on the summed value.
        // Then compute channel average by dividing by the number of spatial elements.
        float channel_avg = (sdata[0] * multiplier) / num_elements;
        atomicAdd(global_accum, channel_avg);
    }
}

// Module function that applies a transposed convolution, fuses elementwise multiplication and global average pooling
at::Tensor fused_module_fn(
    at::Tensor x,
    int64_t stride,
    int64_t padding,
    int64_t output_padding,
    at::Tensor conv_transpose,
    at::Tensor conv_transpose_bias,
    double multiplier
) {
    // Perform transposed convolution (implemented by PyTorch) 
    at::Tensor y = at::conv_transpose2d(
        x,
        conv_transpose,
        conv_transpose_bias,
        {stride, stride},
        {padding, padding},
        {output_padding, output_padding},
        1,
        {1, 1}
    );

    // Instead of doing an out-of-place elementwise multiplication, we fuse the multiplier in our reduction kernel
    
    // Prepare a tensor on the same device to accumulate the global sum
    auto options = torch::TensorOptions().device(y.device()).dtype(y.dtype());
    at::Tensor global_accum_tensor = torch::zeros({1}, options);
    
    // Kernel configuration
    constexpr int BLOCK_SIZE = 256;
    int N = y.size(0);
    int C = y.size(1);
    int H = y.size(2);
    int W = y.size(3);
    int num_channels = N * C; // one block per (batch, channel) combination
    int shared_mem_size = BLOCK_SIZE * sizeof(float);
    
    fused_global_avg_kernel<BLOCK_SIZE><<<num_channels, BLOCK_SIZE, shared_mem_size>>>(
        y.data_ptr<float>(),
        global_accum_tensor.data_ptr<float>(),
        N, C, H, W,
        static_cast<float>(multiplier)
    );

    // Ensure kernel completes
    cudaDeviceSynchronize();

    // The kernel has accumulated the per-channel averages into global_accum_tensor[0].
    // Final result (global average) is the mean of these channel averages.
    at::Tensor result = global_accum_tensor / static_cast<float>(num_channels);
    return result;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &fused_module_fn, "Fused Global Average Pooling Module");
}
